@@ -266,10 +266,18 @@ int Create(struct supertype *st, char *mddev,
 					      &chunk, size*2, NULL, &newsize, verbose>=0))
 		return 1;
 
-	if (chunk) {
+	if (chunk && chunk != UnSet) {
 		newsize &= ~(unsigned long long)(chunk*2 - 1);
-		size &= ~(unsigned long long)(chunk - 1);
+		if (do_default_chunk) {
+			/* default chunk was just set */
+			if (verbose > 0)
+				fprintf(stderr, Name ": chunk size "
+					"defaults to %dK\n", chunk);
+			size &= ~(unsigned long long)(chunk - 1);
+			do_default_chunk = 0;
+		}
 	}
+
 	if (size == 0) {
 		size = newsize / 2;
 		if (size && verbose > 0)
@@ -284,6 +292,8 @@ int Create(struct supertype *st, char *mddev,
 	for (dv=devlist; dv && !have_container; dv=dv->next, dnum++) {
 		char *dname = dv->devname;
 		unsigned long long freesize;
+		int dfd;
+
 		if (strcasecmp(dname, "missing")==0) {
 			if (first_missing > dnum)
 				first_missing = dnum;
@@ -292,6 +302,20 @@ int Create(struct supertype *st, char *mddev,
 			missing_disks ++;
 			continue;
 		}
+		dfd = open(dname, O_RDONLY);
+		if (dfd < 0) {
+			fprintf(stderr, Name ": cannot open %s: %s\n",
+				dname, strerror(errno));
+			exit(2);
+		}
+		if (fstat(dfd, &stb) != 0 ||
+		    (stb.st_mode & S_IFMT) != S_IFBLK) {
+			close(dfd);
+			fprintf(stderr, Name ": %s is not a block device\n",
+				dname);
+			exit(2);
+		}
+		close(dfd);
 		info.array.working_disks++;
 		if (dnum < raiddisks)
 			info.array.active_disks++;
@@ -308,19 +332,35 @@ int Create(struct supertype *st, char *mddev,
 			char *name = "default";
 			for(i=0; !st && superlist[i]; i++) {
 				st = superlist[i]->match_metadata_desc(name);
+				if (!st)
+					continue;
 				if (do_default_layout)
 					layout = default_layout(st, level, verbose);
-				if (st && !st->ss->validate_geometry
-					    	(st, level, layout, raiddisks,
-						 &chunk, size*2, dname, &freesize,
-						 verbose > 0)) {
+				switch (st->ss->validate_geometry(
+						st, level, layout, raiddisks,
+						&chunk, size*2, dname, &freesize,
+						verbose > 0)) {
+				case -1: /* Not valid, message printed, and not
+					  * worth checking any further */
+					exit(2);
+					break;
+				case 0: /* Geometry not valid */
 					free(st);
 					st = NULL;
-					chunk = do_default_chunk ? 0 : chunk;
+					chunk = do_default_chunk ? UnSet : chunk;
+					break;
+				case 1:	/* All happy */
+					break;
 				}
 			}
 
 			if (!st) {
+				int dfd = open(dname, O_RDONLY|O_EXCL);
+				if (dfd < 0) {
+					fprintf(stderr, Name ": cannot open %s: %s\n",
+						dname, strerror(errno));
+					exit(2);
+				}
 				fprintf(stderr, Name ": device %s not suitable "
 					"for any style of array\n",
 					dname);
@@ -346,16 +386,19 @@ int Create(struct supertype *st, char *mddev,
 				continue;
 			}
 		}
-		if (verbose > 0 && do_default_chunk) {
-			do_default_chunk = 0;
-			fprintf(stderr, Name ": chunk size "
-				"defaults to %dK\n", chunk);
-		}
 
 		freesize /= 2; /* convert to K */
-		if (chunk) {
+		if (chunk && chunk != UnSet) {
 			/* round to chunk size */
 			freesize = freesize & ~(chunk-1);
+			if (do_default_chunk) {
+				/* default chunk was just set */
+				if (verbose > 0)
+					fprintf(stderr, Name ": chunk size "
+						"defaults to %dK\n", chunk);
+				size &= ~(unsigned long long)(chunk - 1);
+				do_default_chunk = 0;
+			}
 		}
 
 		if (size && freesize < size) {
@@ -511,14 +554,28 @@ int Create(struct supertype *st, char *mddev,
 	/* We need to create the device */
 	map_lock(&map);
 	mdfd = create_mddev(mddev, name, autof, LOCAL, chosen_name);
-	if (mdfd < 0)
+	if (mdfd < 0) {
+		map_unlock(&map);
 		return 1;
+	}
+	/* verify if chosen_name is not in use,
+	 * it could be in conflict with already existing device
+	 * e.g. container, array
+	 */
+	if (strncmp(chosen_name, "/dev/md/", 8) == 0
+	    && map_by_name(&map, chosen_name+8) != NULL) {
+		fprintf(stderr, Name ": Array name %s is in use already.\n",
+			chosen_name);
+		close(mdfd);
+		map_unlock(&map);
+		return 1;
+	}
 	mddev = chosen_name;
 
 	vers = md_get_version(mdfd);
 	if (vers < 9000) {
 		fprintf(stderr, Name ": Create requires md driver version 0.90.0 or later\n");
-		goto abort;
+		goto abort_locked;
 	} else {
 		mdu_array_info_t inf;
 		memset(&inf, 0, sizeof(inf));
@@ -526,7 +583,7 @@ int Create(struct supertype *st, char *mddev,
 		if (inf.working_disks != 0) {
 			fprintf(stderr, Name ": another array by this name"
 				" is already running.\n");
-			goto abort;
+			goto abort_locked;
 		}
 	}
 
@@ -622,11 +679,11 @@ int Create(struct supertype *st, char *mddev,
 		}
 	}
 	if (!st->ss->init_super(st, &info.array, size, name, homehost, uuid))
-		goto abort;
+		goto abort_locked;
 
 	total_slots = info.array.nr_disks;
-	sysfs_init(&info, mdfd, 0);
 	st->ss->getinfo_super(st, &info, NULL);
+	sysfs_init(&info, mdfd, 0);
 
 	if (did_default && verbose >= 0) {
 		if (is_subarray(info.text_version)) {
@@ -745,6 +802,10 @@ int Create(struct supertype *st, char *mddev,
 	}
 
 	infos = malloc(sizeof(*infos) * total_slots);
+	if (!infos) {
+		fprintf(stderr, Name ": Unable to allocate memory\n");
+		goto abort;
+	}
 
 	for (pass=1; pass <=2 ; pass++) {
 		struct mddev_dev *moved_disk = NULL; /* the disk that was moved out of the insert point */
@@ -827,7 +888,6 @@ int Create(struct supertype *st, char *mddev,
 				break;
 			case 2:
 				inf->errors = 0;
-				rv = 0;
 
 				rv = add_disk(mdfd, st, &info, inf);
 
@@ -863,7 +923,13 @@ int Create(struct supertype *st, char *mddev,
 				me = map_by_devnum(&map, st->container_dev);
 			}
 
-			st->ss->write_init_super(st);
+			if (st->ss->write_init_super(st)) {
+				fprintf(stderr,
+					Name ": Failed to write metadata to %s\n",
+					dv->devname);
+				st->ss->free_super(st);
+				goto abort_locked;
+			}
 
 			/* update parent container uuid */
 			if (me) {
@@ -919,6 +985,10 @@ int Create(struct supertype *st, char *mddev,
 			if (ioctl(mdfd, RUN_ARRAY, &param)) {
 				fprintf(stderr, Name ": RUN_ARRAY failed: %s\n",
 					strerror(errno));
+				if (info.array.chunk_size & (info.array.chunk_size-1)) {
+					fprintf(stderr, "     : Problem may be that "
+						"chunk size is not a power of 2\n");
+				}
 				ioctl(mdfd, STOP_ARRAY, NULL);
 				goto abort;
 			}
@@ -941,6 +1011,7 @@ int Create(struct supertype *st, char *mddev,
 
  abort:
 	map_lock(&map);
+ abort_locked:
 	map_remove(&map, fd2devnum(mdfd));
 	map_unlock(&map);
 
