@@ -1,8 +1,8 @@
 /*
- * mapfile - manage /var/run/mdadm.map. Part of:
+ * mapfile - keep track of uuid <-> array mapping. Part of:
  * mdadm - manage Linux "md" devices aka RAID arrays.
  *
- * Copyright (C) 2006-2009 Neil Brown <neilb@suse.de>
+ * Copyright (C) 2006-2010 Neil Brown <neilb@suse.de>
  *
  *
  *    This program is free software; you can redistribute it and/or modify
@@ -28,8 +28,8 @@
  *           Australia
  */
 
-/* /var/run/mdadm.map is used to track arrays being created in --incremental
- * more.  It particularly allows lookup from UUID to array device, but
+/* The mapfile is used to track arrays being created in --incremental
+ * mode.  It particularly allows lookup from UUID to array device, but
  * also allows the array device name to be easily found.
  *
  * The map file is line based with space separated fields.  The fields are:
@@ -38,38 +38,49 @@
  *  UUID       -  uuid of the array
  *  path       -  path where device created: /dev/md/home
  *
- * The preferred location for the map file is /var/run/mdadm.map.
- * However /var/run may not exist or be writable in early boot.  And if
- * no-one has created /var/run/mdadm, we still want to survive.
- * So possible locations are:
- *   /var/run/mdadm/map  /var/run/mdadm.map  /dev/.mdadm.map
- * the last, because udev requires a writable /dev very early.
- * We read from the first one that exists and write to the first
- * one that we can.
+ * The best place for the mapfile wold be /var/run/mdadm/map.  However
+ * it is needed during initramfs early-boot, and /var/run doesn't exist there
+ * and certainly doesn't persist through to normal boot.
+ * So we store it in /dev/.mdadm/map but allow this to be changed at
+ * compile time. via MAP_DIR and MAP_FILE
+ *
  */
 #include	"mdadm.h"
+#include	<sys/file.h>
 #include	<ctype.h>
 
-#define mapnames(base) { #base, #base ".new", #base ".lock"}
-char *mapname[3][3] = {
-	mapnames(/var/run/mdadm/map),
-	mapnames(/var/run/mdadm.map),
-	mapnames(/dev/.mdadm.map)
+#ifndef MAP_DIR
+#define MAP_DIR "/dev/.mdadm"
+#define MAP_FILE "map"
+#endif
+
+#define MAP_READ 0
+#define MAP_NEW 1
+#define MAP_LOCK 2
+#define MAP_DIRNAME 3
+#define mapnames(dir, base) { \
+
+char *mapname[4] = {
+	MAP_DIR "/" MAP_FILE,
+	MAP_DIR "/" MAP_FILE ".new",
+	MAP_DIR "/" MAP_FILE ".lock",
+	MAP_DIR
 };
 
-int mapmode[3] = { O_RDONLY, O_RDWR|O_CREAT, O_RDWR|O_CREAT | O_TRUNC };
+int mapmode[3] = { O_RDONLY, O_RDWR|O_CREAT, O_RDWR|O_CREAT|O_TRUNC };
 char *mapsmode[3] = { "r", "w", "w"};
 
-FILE *open_map(int modenum, int *choice)
+FILE *open_map(int modenum)
 {
-	int i;
-	for (i = 0 ; i < 3 ; i++) {
-		int fd = open(mapname[i][modenum], mapmode[modenum], 0600);
-		if (fd >= 0) {
-			*choice = i;
-			return fdopen(fd, mapsmode[modenum]);
-		}
-	}
+	int fd;
+	if ((mapmode[modenum] & O_CREAT))
+		/* Attempt to create directory, don't worry about
+		 * failure.
+		 */
+		(void)mkdir(mapname[MAP_DIRNAME], 0755);
+	fd = open(mapname[modenum], mapmode[modenum], 0600);
+	if (fd >= 0)
+		return fdopen(fd, mapsmode[modenum]);
 	return NULL;
 }
 
@@ -77,9 +88,8 @@ int map_write(struct map_ent *mel)
 {
 	FILE *f;
 	int err;
-	int which;
 
-	f = open_map(1, &which);
+	f = open_map(MAP_NEW);
 
 	if (!f)
 		return 0;
@@ -99,27 +109,36 @@ int map_write(struct map_ent *mel)
 	err = ferror(f);
 	fclose(f);
 	if (err) {
-		unlink(mapname[which][1]);
+		unlink(mapname[1]);
 		return 0;
 	}
-	return rename(mapname[which][1],
-		      mapname[which][0]) == 0;
+	return rename(mapname[1],
+		      mapname[0]) == 0;
 }
 
 
 static FILE *lf = NULL;
-static int lwhich = 0;
 int map_lock(struct map_ent **melp)
 {
-	if (lf == NULL) {
-		lf = open_map(2, &lwhich);
+	while (lf == NULL) {
+		struct stat buf;
+		lf = open_map(MAP_LOCK);
 		if (lf == NULL)
 			return -1;
-		if (lockf(fileno(lf), F_LOCK, 0) != 0) {
+		if (flock(fileno(lf), LOCK_EX) != 0) {
 			fclose(lf);
 			lf = NULL;
 			return -1;
 		}
+		if (fstat(fileno(lf), &buf) != 0 ||
+		    buf.st_nlink == 0) {
+			/* The owner of the lock unlinked it,
+			 * so we have a lock on a stale file,
+			 * try again
+			 */
+			fclose(lf);
+			lf = NULL;
+		}
 	}
 	if (*melp)
 		map_free(*melp);
@@ -129,12 +148,29 @@ int map_lock(struct map_ent **melp)
 
 void map_unlock(struct map_ent **melp)
 {
-	if (lf)
+	if (lf) {
+		/* must unlink before closing the file,
+		 * as only the owner of the lock may
+		 * unlink the file
+		 */
+		unlink(mapname[2]);
 		fclose(lf);
-	unlink(mapname[lwhich][2]);
+	}
 	lf = NULL;
 }
 
+void map_fork(void)
+{
+	/* We are forking, so must close the lock file.
+	 * Don't risk flushing anything though.
+	 */
+	if (lf) {
+		close(fileno(lf));
+		fclose(lf);
+		lf = NULL;
+	}
+}
+
 void map_add(struct map_ent **melp,
 	    int devnum, char *metadata, int uuid[4], char *path)
 {
@@ -157,14 +193,13 @@ void map_read(struct map_ent **melp)
 	int devnum, uuid[4];
 	char metadata[30];
 	char nam[4];
-	int which;
 
 	*melp = NULL;
 
-	f = open_map(0, &which);
+	f = open_map(MAP_READ);
 	if (!f) {
 		RebuildMap();
-		f = open_map(0, &which);
+		f = open_map(MAP_READ);
 	}
 	if (!f)
 		return;
@@ -239,6 +274,16 @@ void map_delete(struct map_ent **mapp, int devnum)
 	}
 }
 
+void map_remove(struct map_ent **mapp, int devnum)
+{
+	if (devnum == NoMdDev)
+		return;
+
+	map_delete(mapp, devnum);
+	map_write(*mapp);
+	map_free(*mapp);
+}
+
 struct map_ent *map_by_uuid(struct map_ent **map, int uuid[4])
 {
 	struct map_ent *mp;
@@ -301,31 +346,20 @@ struct map_ent *map_by_name(struct map_ent **map, char *name)
  * version super_by_fd does this automatically, this routine is meant as
  * a supplement for guess_super()
  */
-static void set_member_info(struct supertype *st, struct mdstat_ent *ent)
+static char *get_member_info(struct mdstat_ent *ent)
 {
 
-	st->subarray[0] = '\0';
-
 	if (ent->metadata_version == NULL ||
 	    strncmp(ent->metadata_version, "external:", 9) != 0)
-		return;
+		return NULL;
 
 	if (is_subarray(&ent->metadata_version[9])) {
-		char version[strlen(ent->metadata_version)+1];
 		char *subarray;
-		char *name = &version[10];
 
-		strcpy(version, ent->metadata_version);
-		subarray = strrchr(version, '/');
-		name = &version[10];
-
-		if (!subarray)
-			return;
-		*subarray++ = '\0';
-
-		st->container_dev = devname2devnum(name);
-		strncpy(st->subarray, subarray, sizeof(st->subarray));
+		subarray = strrchr(ent->metadata_version, '/');
+		return subarray + 1;
 	}
+	return NULL;
 }
 
 void RebuildMap(void)
@@ -346,7 +380,7 @@ void RebuildMap(void)
 	}
 
 	for (md = mdstat ; md ; md = md->next) {
-		struct mdinfo *sra = sysfs_read(-1, md->devnum, GET_DEVS|SKIP_GONE_DEVS);
+		struct mdinfo *sra = sysfs_read(-1, md->devnum, GET_DEVS);
 		struct mdinfo *sd;
 
 		if (!sra)
@@ -358,8 +392,9 @@ void RebuildMap(void)
 			int dfd;
 			int ok;
 			struct supertype *st;
+			char *subarray = NULL;
 			char *path;
-			struct mdinfo info;
+			struct mdinfo *info;
 
 			sprintf(dn, "%d:%d", sd->disk.major, sd->disk.minor);
 			dfd = dev_open(dn, O_RDONLY);
@@ -369,13 +404,14 @@ void RebuildMap(void)
 			if ( st == NULL)
 				ok = -1;
 			else {
-				set_member_info(st, md);
+				subarray = get_member_info(md);
 				ok = st->ss->load_super(st, dfd, NULL);
 			}
 			close(dfd);
 			if (ok != 0)
 				continue;
-			st->ss->getinfo_super(st, &info);
+			info = st->ss->container_content(st, subarray);
+
 			if (md->devnum >= 0)
 				path = map_dev(MD_MAJOR, md->devnum, 0);
 			else
@@ -395,7 +431,9 @@ void RebuildMap(void)
 				 *   find a unique name based on metadata name.
 				 *   
 				 */
-				struct mddev_ident_s *match = conf_match(&info, st);
+				struct mddev_ident *match = conf_match(st, info,
+								       NULL, 0,
+								       NULL);
 				struct stat stb;
 				if (match && match->devname && match->devname[0] == '/') {
 					path = match->devname;
@@ -413,13 +451,13 @@ void RebuildMap(void)
 					     st->ss->match_home(st, homehost) != 1) &&
 					    st->ss->match_home(st, "any") != 1 &&
 					    (require_homehost
-					     || ! conf_name_is_free(info.name)))
+					     || ! conf_name_is_free(info->name)))
 						/* require a numeric suffix */
 						unum = 0;
 					else
 						/* allow name to be used as-is if no conflict */
 						unum = -1;
-					name = info.name;
+					name = info->name;
 					if (!*name) {
 						name = st->ss->name;
 						if (!isdigit(name[strlen(name)-1]) &&
@@ -452,19 +490,23 @@ void RebuildMap(void)
 				}
 			}
 			map_add(&map, md->devnum,
-				info.text_version,
-				info.uuid, path);
+				info->text_version,
+				info->uuid, path);
 			st->ss->free_super(st);
+			free(info);
 			break;
 		}
 		sysfs_free(sra);
 	}
-	map_write(map);
+	/* Only trigger a change if we wrote a new map file */
+	if (map_write(map))
+		for (md = mdstat ; md ; md = md->next) {
+			struct mdinfo *sra = sysfs_read(-1, md->devnum,
+							GET_VERSION);
+			if (sra)
+				sysfs_uevent(sra, "change");
+			sysfs_free(sra);
+		}
 	map_free(map);
-	for (md = mdstat ; md ; md = md->next) {
-		struct mdinfo *sra = sysfs_read(-1, md->devnum, GET_VERSION);
-		sysfs_uevent(sra, "change");
-		sysfs_free(sra);
-	}
 	free_mdstat(mdstat);
 }
